/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package vcl

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"path/filepath"
	"runtime"
	"strings"
	"testing"
	"text/template"
	"time"
)

var (
	testdir string
	tmpdir  string
)

func cmpGold(t *testing.T, gotPath string, goldfile string) bool {
	goldPath := filepath.Join(testdir, "testdata", goldfile)
	gold, err := ioutil.ReadFile(goldPath)
	if err != nil {
		t.Fatalf("Cannot read %s: %s", goldPath, err)
		return false
	}
	got, err := ioutil.ReadFile(gotPath)
	if err != nil {
		t.Fatalf("Cannot read %s: %s", gotPath, err)
		return false
	}
	return bytes.Equal(got, gold)
}

func TestMain(m *testing.M) {
	var err error

	_, filename, _, ok := runtime.Caller(0)
	if !ok {
		fmt.Println("Cannot get test directory")
		os.Exit(-1)
	}
	testdir = filepath.Dir(filename)
	tmpdir, err = ioutil.TempDir("", "k8s-vcl-reloader")
	if err != nil {
		fmt.Println("Cannot create tmp directory:", err)
		os.Exit(-1)
	}
	tmpdir = filepath.Join(tmpdir, "vcl_test")
	if err = os.Mkdir(tmpdir, 0755); err != nil {
		fmt.Printf("Cannot make tmp directory %s: %s", tmpdir, err)
		os.Exit(-1)
	}
	code := m.Run()
	os.Exit(code)
}

var testSpec = Spec{
	Version: "4.7.11",
	Created: time.Unix(1136239445, 123456000),
	Main:    "default.vcl",
	CfgMap: ConfigMap{
		Name:    "vcl-cfgmap",
		Version: "4.7.11",
		Meta: Meta{
			Namespace:       "default",
			UID:             "513e6391-079b-11ea-b351-bab55b1b33a8",
			ResourceVersion: "520822",
		},
	},
	Services: map[string]Service{
		"bar-service": {
			Meta: Meta{
				Namespace:       "default",
				UID:             "47850e08-0791-11ea-b351-bab55b1b33a8",
				ResourceVersion: "514058",
			},
			Endpoints: map[string]Endpoint{
				"bar-64548cd7b6-j5b5w": {
					Meta: Meta{
						Namespace:       "default",
						UID:             "d7d27ba8-0792-11ea-b351-bab55b1b33a8",
						ResourceVersion: "515214",
					},
					IP:   "172.22.225.13",
					Port: 8080,
				},
				"bar-64548cd7b6-fb2gb": {
					Meta: Meta{
						Namespace:       "default",
						UID:             "d52a8e5b-0792-11ea-b351-bab55b1b33a8",
						ResourceVersion: "515183",
					},
					IP:   "172.22.225.49",
					Port: 8080,
				},
			},
		},
		"baz-service": {
			Meta: Meta{
				Namespace:       "default",
				UID:             "993323b2-077f-11ea-b351-bab55b1b33a8",
				ResourceVersion: "502259",
			},
			Endpoints: map[string]Endpoint{
				"baz-5ff49fcd4-hx7lj": {
					Meta: Meta{
						Namespace:       "default",
						UID:             "efdb0528-0792-11ea-b351-bab55b1b33a8",
						ResourceVersion: "515894",
					},
					IP:   "172.22.225.8",
					Port: 8080,
				},
			},
		},
		"foo-service": {
			Meta: Meta{
				Namespace:       "default",
				UID:             "5145d363-079b-11ea-b351-bab55b1b33a8",
				ResourceVersion: "520824",
			},
			Endpoints: map[string]Endpoint{
				"foo-stateful-0": {
					Meta: Meta{
						Namespace:       "default",
						UID:             "51712506-079b-11ea-b351-bab55b1b33a8",
						ResourceVersion: "520869",
					},
					IP:   "172.22.225.25",
					Port: 6081,
				},
				"foo-stateful-1": {
					Meta: Meta{
						Namespace:       "default",
						UID:             "56f812d1-079b-11ea-b351-bab55b1b33a8",
						ResourceVersion: "520898",
					},
					IP:   "172.22.225.39",
					Port: 6081,
				},
			},
		},
	},
}

func TestTemplate(t *testing.T) {
	var tmpl *template.Template

	gold := "backends.golden"
	tmplMap := make(map[string]string)
	vclPath := filepath.Join(tmpdir, "backends.vcl")
	tmplMap[filepath.Join(testdir, "testdata", "backends.tmpl")] = vclPath

	tmpl2file, err := InitTemplates(tmplMap)
	if err != nil {
		t.Fatal("InitTemplates():", err)
	}
	for t := range tmpl2file {
		tmpl = t
	}

	if err = testSpec.WriteTemplate(tmpl, vclPath); err != nil {
		t.Fatal("WriteTemplate():", err)
	}
	if !cmpGold(t, vclPath, gold) {
		t.Errorf("Generated VCL at %s does not match gold file %s",
			vclPath, filepath.Join("testdata", gold))
	}

	gold = "backends_nobar.golden"
	noEpsSpec := testSpec
	svc, _ := noEpsSpec.Services["bar-service"]
	svc.Endpoints = map[string]Endpoint{}
	noEpsSpec.Services["bar-service"] = svc
	if err = noEpsSpec.WriteTemplate(tmpl, vclPath); err != nil {
		t.Fatal("WriteTemplate():", err)
	}
	if !cmpGold(t, vclPath, gold) {
		t.Errorf("Generated VCL at %s does not match gold file %s",
			vclPath, filepath.Join("testdata", gold))
	}

	gold = "backends_nosvc.golden"
	svc, _ = noEpsSpec.Services["baz-service"]
	svc.Endpoints = map[string]Endpoint{}
	noEpsSpec.Services["baz-service"] = svc
	if err = noEpsSpec.WriteTemplate(tmpl, vclPath); err != nil {
		t.Fatal("WriteTemplate():", err)
	}
	if !cmpGold(t, vclPath, gold) {
		t.Errorf("Generated VCL at %s does not match gold file %s",
			vclPath, filepath.Join("testdata", gold))
	}
}

func TestFQVersion(t *testing.T) {
	fqVersion := testSpec.FQVersion()
	if testing.Verbose() {
		t.Log("FQVersion():", fqVersion)
	}
	if !strings.HasPrefix(fqVersion,
		testSpec.CfgMap.Namespace+"/"+testSpec.CfgMap.Name+"-") {
		t.Fatal("FQVersion(): does not begin with ConfigMap name + "+
			"\"-\":", fqVersion)
	}
	versionSlice := strings.SplitN(fqVersion, "-", 4)
	if versionSlice[2] != testSpec.Version {
		t.Fatalf("FQVersion(): does not include spec.Version=%s (%s): "+
			"%s", testSpec.Version, versionSlice[2], fqVersion)
	}
	if created, err := time.Parse(RFC3339Micro, versionSlice[3]); err != nil {
		t.Fatal("FQVersion(): cannot parse timestamp", err)
	} else if !testSpec.Created.Equal(created) {
		t.Fatalf("FQVersion(): timestamp != Spec.Created (%s) "+
			"got %s want %s", versionSlice[3], testSpec.Created,
			created)
	}

	noversionSpec := Spec{
		Created: time.Unix(1136239445, 123456000),
		CfgMap: ConfigMap{
			Name: "vcl-cfgmap",
			Meta: Meta{
				Namespace: "default",
			},
		},
	}
	fqVersion = noversionSpec.FQVersion()
	if testing.Verbose() {
		t.Log("FQVersion():", fqVersion)
	}
	versionSlice = strings.SplitN(fqVersion, "-", 3)
	if created, err := time.Parse(RFC3339Micro, versionSlice[2]); err != nil {
		t.Fatal("FQVersion(): cannot parse timestamp", err)
	} else if !noversionSpec.Created.Equal(created) {
		t.Fatal("FQVersion(): timestamp != Spec.Created", fqVersion)
	}
}

func TestConfigName(t *testing.T) {
	cfgName := testSpec.ConfigName()
	if testing.Verbose() {
		t.Log("ConfigName():", cfgName)
	}
	if vclIllegal.MatchString(cfgName) {
		t.Fatal("ConfigName() result is an illegal VCL name:", cfgName)
	}
}

func TestDeepHash(t *testing.T) {
	specCopy := testSpec
	if !bytes.Equal(specCopy.DeepHash(), testSpec.DeepHash()) {
		t.Fatalf("DeepHash(%+v)=%+v != DeepHash(%+v)=%+v", testSpec,
			testSpec.DeepHash(), specCopy, specCopy.DeepHash())
	}
}
